# -*-*- encoding: utf-8 -*-*-
#
# This is the base file for the PicasaWeb python client.
# It is used for lower level operations.
#
# $Id: __init__.py 148 2007-10-28 15:09:19Z havard.gulldahl $
#
# Copyright 2007 Håvard Gulldahl 
# Portions (C) 2006 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides a pythonic, gdata-centric interface to Google Photos
(a.k.a. Picasa Web Services.

It is modelled after the gdata/* interfaces from the gdata-python-client
project[1] by Google. 

You'll find the user-friendly api in photos.service. Please see the
documentation or live help() system for available methods.

[1]: http://gdata-python-client.googlecode.com/

  """

__author__ = u'havard@gulldahl.no'# (Håvard Gulldahl)' #BUG: pydoc chokes on non-ascii chars in __author__
__license__ = 'Apache License v2'
__version__ = '$Revision: 164 $'[11:-2]

import re
try:
  from xml.etree import cElementTree as ElementTree
except ImportError:
  try:
    import cElementTree as ElementTree
  except ImportError:
    try:
      from xml.etree import ElementTree
    except ImportError:
      from elementtree import ElementTree
import atom
import gdata

# importing google photo submodules
import gdata.media as Media, gdata.exif as Exif, gdata.geo as Geo

# XML namespaces which are often used in Google Photo elements
PHOTOS_NAMESPACE = 'http://schemas.google.com/photos/2007'
MEDIA_NAMESPACE = 'http://search.yahoo.com/mrss/'
EXIF_NAMESPACE = 'http://schemas.google.com/photos/exif/2007'
OPENSEARCH_NAMESPACE = 'http://a9.com/-/spec/opensearchrss/1.0/'
GEO_NAMESPACE = 'http://www.w3.org/2003/01/geo/wgs84_pos#'
GML_NAMESPACE = 'http://www.opengis.net/gml'
GEORSS_NAMESPACE = 'http://www.georss.org/georss'
PHEED_NAMESPACE = 'http://www.pheed.com/pheed/'
BATCH_NAMESPACE = 'http://schemas.google.com/gdata/batch'


class PhotosBaseElement(atom.AtomBase):
  """Base class for elements in the PHOTO_NAMESPACE. To add new elements,
  you only need to add the element tag name to self._tag
  """
  
  _tag = ''
  _namespace = PHOTOS_NAMESPACE
  _children = atom.AtomBase._children.copy()
  _attributes = atom.AtomBase._attributes.copy()

  def __init__(self, name=None, extension_elements=None,
      extension_attributes=None, text=None):
    self.name = name
    self.text = text
    self.extension_elements = extension_elements or []
    self.extension_attributes = extension_attributes or {}
  #def __str__(self):
    #return str(self.text)
  #def __unicode__(self):
    #return unicode(self.text)
  def __int__(self):
    return int(self.text)
  def bool(self):
    return self.text == 'true'

class GPhotosBaseFeed(gdata.GDataFeed, gdata.LinkFinder):
  "Base class for all Feeds in gdata.photos"
  _tag = 'feed'
  _namespace = atom.ATOM_NAMESPACE
  _attributes = gdata.GDataFeed._attributes.copy()
  _children = gdata.GDataFeed._children.copy()
  # We deal with Entry elements ourselves
  del _children['{%s}entry' % atom.ATOM_NAMESPACE]
    
  def __init__(self, author=None, category=None, contributor=None,
               generator=None, icon=None, atom_id=None, link=None, logo=None,
               rights=None, subtitle=None, title=None, updated=None,
               entry=None, total_results=None, start_index=None,
               items_per_page=None, extension_elements=None,
               extension_attributes=None, text=None):
    gdata.GDataFeed.__init__(self, author=author, category=category,
                             contributor=contributor, generator=generator,
                             icon=icon,  atom_id=atom_id, link=link,
                             logo=logo, rights=rights, subtitle=subtitle,
                             title=title, updated=updated, entry=entry,
                             total_results=total_results,
                             start_index=start_index,
                             items_per_page=items_per_page,
                             extension_elements=extension_elements,
                             extension_attributes=extension_attributes,
                             text=text)

  def kind(self):
    "(string) Returns the kind"
    try:
      return self.category[0].term.split('#')[1]
    except IndexError:
      return None
  
  def _feedUri(self, kind):
    "Convenience method to return a uri to a feed of a special kind"
    assert(kind in ('album', 'tag', 'photo', 'comment', 'user'))
    here_href = self.GetSelfLink().href
    if 'kind=%s' % kind in here_href:
      return here_href
    if not 'kind=' in here_href:
      sep = '?'
      if '?' in here_href: sep = '&'
      return here_href + "%skind=%s" % (sep, kind)
    rx = re.match('.*(kind=)(album|tag|photo|comment)', here_href)
    return here_href[:rx.end(1)] + kind + here_href[rx.end(2):]
  
  def _ConvertElementTreeToMember(self, child_tree):
    """Re-implementing the method from AtomBase, since we deal with
      Entry elements specially"""
    category = child_tree.find('{%s}category' % atom.ATOM_NAMESPACE)
    if category is None:
      return atom.AtomBase._ConvertElementTreeToMember(self, child_tree)
    namespace, kind = category.get('term').split('#')
    if namespace != PHOTOS_NAMESPACE:
      return atom.AtomBase._ConvertElementTreeToMember(self, child_tree)
    ## TODO: is it safe to use getattr on gdata.photos?
    entry_class = getattr(gdata.photos, '%sEntry' % kind.title())
    if not hasattr(self, 'entry') or self.entry is None:
      self.entry = []
    self.entry.append(atom._CreateClassFromElementTree(
        entry_class, child_tree))

class GPhotosBaseEntry(gdata.GDataEntry, gdata.LinkFinder):
  "Base class for all Entry elements in gdata.photos"
  _tag = 'entry'
  _kind = ''
  _namespace = atom.ATOM_NAMESPACE
  _children = gdata.GDataEntry._children.copy()
  _attributes = gdata.GDataEntry._attributes.copy()
    
  def __init__(self, author=None, category=None, content=None,
      atom_id=None, link=None, published=None,
      title=None, updated=None,
      extended_property=None,
      extension_elements=None, extension_attributes=None, text=None):
    gdata.GDataEntry.__init__(self, author=author, category=category,
                        content=content, atom_id=atom_id, link=link,
                        published=published, title=title,
                        updated=updated, text=text,
                        extension_elements=extension_elements,
                        extension_attributes=extension_attributes)
    self.category.append(
      atom.Category(scheme='http://schemas.google.com/g/2005#kind', 
              term = 'http://schemas.google.com/photos/2007#%s' % self._kind))

  def kind(self):
    "(string) Returns the kind"
    try:
      return self.category[0].term.split('#')[1]
    except IndexError:
      return None
  
  def _feedUri(self, kind):
    "Convenience method to get the uri to this entry's feed of the some kind"
    try:
      href = self.GetFeedLink().href
    except AttributeError:
      return None
    sep = '?'
    if '?' in href: sep = '&'
    return '%s%skind=%s' % (href, sep, kind)


class PhotosBaseEntry(GPhotosBaseEntry):
  pass

class PhotosBaseFeed(GPhotosBaseFeed):
  pass

class GPhotosBaseData(object):
  pass

class Access(PhotosBaseElement):
  """The Google Photo `Access' element.

  The album's access level. Valid values are `public' or `private'.
  In documentation, access level is also referred to as `visibility.'"""
  
  _tag = 'access'
def AccessFromString(xml_string):
  return atom.CreateClassFromXMLString(Access, xml_string)

class Albumid(PhotosBaseElement):
  "The Google Photo `Albumid' element"
  
  _tag = 'albumid'
def AlbumidFromString(xml_string):
  return atom.CreateClassFromXMLString(Albumid, xml_string)

class BytesUsed(PhotosBaseElement):
  "The Google Photo `BytesUsed' element"
  
  _tag = 'bytesUsed'
def BytesUsedFromString(xml_string):
  return atom.CreateClassFromXMLString(BytesUsed, xml_string)

class Client(PhotosBaseElement):
  "The Google Photo `Client' element"
  
  _tag = 'client'
def ClientFromString(xml_string):
  return atom.CreateClassFromXMLString(Client, xml_string)

class Checksum(PhotosBaseElement):
  "The Google Photo `Checksum' element"
  
  _tag = 'checksum'
def ChecksumFromString(xml_string):
  return atom.CreateClassFromXMLString(Checksum, xml_string)

class CommentCount(PhotosBaseElement):
  "The Google Photo `CommentCount' element"
  
  _tag = 'commentCount'
def CommentCountFromString(xml_string):
  return atom.CreateClassFromXMLString(CommentCount, xml_string)

class CommentingEnabled(PhotosBaseElement):
  "The Google Photo `CommentingEnabled' element"
  
  _tag = 'commentingEnabled'
def CommentingEnabledFromString(xml_string):
  return atom.CreateClassFromXMLString(CommentingEnabled, xml_string)

class Height(PhotosBaseElement):
  "The Google Photo `Height' element"
  
  _tag = 'height'
def HeightFromString(xml_string):
  return atom.CreateClassFromXMLString(Height, xml_string)

class Id(PhotosBaseElement):
  "The Google Photo `Id' element"
  
  _tag = 'id'
def IdFromString(xml_string):
  return atom.CreateClassFromXMLString(Id, xml_string)

class Location(PhotosBaseElement):
  "The Google Photo `Location' element"
  
  _tag = 'location'
def LocationFromString(xml_string):
  return atom.CreateClassFromXMLString(Location, xml_string)

class MaxPhotosPerAlbum(PhotosBaseElement):
  "The Google Photo `MaxPhotosPerAlbum' element"
  
  _tag = 'maxPhotosPerAlbum'
def MaxPhotosPerAlbumFromString(xml_string):
  return atom.CreateClassFromXMLString(MaxPhotosPerAlbum, xml_string)

class Name(PhotosBaseElement):
  "The Google Photo `Name' element"
  
  _tag = 'name'
def NameFromString(xml_string):
  return atom.CreateClassFromXMLString(Name, xml_string)

class Nickname(PhotosBaseElement):
  "The Google Photo `Nickname' element"
  
  _tag = 'nickname'
def NicknameFromString(xml_string):
  return atom.CreateClassFromXMLString(Nickname, xml_string)

class Numphotos(PhotosBaseElement):
  "The Google Photo `Numphotos' element"
  
  _tag = 'numphotos'
def NumphotosFromString(xml_string):
  return atom.CreateClassFromXMLString(Numphotos, xml_string)

class Numphotosremaining(PhotosBaseElement):
  "The Google Photo `Numphotosremaining' element"
  
  _tag = 'numphotosremaining'
def NumphotosremainingFromString(xml_string):
  return atom.CreateClassFromXMLString(Numphotosremaining, xml_string)

class Position(PhotosBaseElement):
  "The Google Photo `Position' element"
  
  _tag = 'position'
def PositionFromString(xml_string):
  return atom.CreateClassFromXMLString(Position, xml_string)

class Photoid(PhotosBaseElement):
  "The Google Photo `Photoid' element"
  
  _tag = 'photoid'
def PhotoidFromString(xml_string):
  return atom.CreateClassFromXMLString(Photoid, xml_string)

class Quotacurrent(PhotosBaseElement):
  "The Google Photo `Quotacurrent' element"
  
  _tag = 'quotacurrent'
def QuotacurrentFromString(xml_string):
  return atom.CreateClassFromXMLString(Quotacurrent, xml_string)

class Quotalimit(PhotosBaseElement):
  "The Google Photo `Quotalimit' element"
  
  _tag = 'quotalimit'
def QuotalimitFromString(xml_string):
  return atom.CreateClassFromXMLString(Quotalimit, xml_string)

class Rotation(PhotosBaseElement):
  "The Google Photo `Rotation' element"
  
  _tag = 'rotation'
def RotationFromString(xml_string):
  return atom.CreateClassFromXMLString(Rotation, xml_string)

class Size(PhotosBaseElement):
  "The Google Photo `Size' element"
  
  _tag = 'size'
def SizeFromString(xml_string):
  return atom.CreateClassFromXMLString(Size, xml_string)

class Snippet(PhotosBaseElement):
  """The Google Photo `snippet' element.

  When searching, the snippet element will contain a 
  string with the word you're looking for, highlighted in html markup
  E.g. when your query is `hafjell', this element may contain:
  `... here at <b>Hafjell</b>.'

  You'll find this element in searches -- that is, feeds that combine the 
  `kind=photo' and `q=yoursearch' parameters in the request.

  See also gphoto:truncated and gphoto:snippettype.
  
  """
  
  _tag = 'snippet'
def SnippetFromString(xml_string):
  return atom.CreateClassFromXMLString(Snippet, xml_string)

class Snippettype(PhotosBaseElement):
  """The Google Photo `Snippettype' element

  When searching, this element will tell you the type of element that matches.

  You'll find this element in searches -- that is, feeds that combine the 
  `kind=photo' and `q=yoursearch' parameters in the request.

  See also gphoto:snippet and gphoto:truncated.
  
  Possible values and their interpretation: 
  o ALBUM_TITLE       - The album title matches 
  o PHOTO_TAGS        - The match is a tag/keyword
  o PHOTO_DESCRIPTION - The match is in the photo's description

  If you discover a value not listed here, please submit a patch to update this docstring.
  
  """
  
  _tag = 'snippettype'
def SnippettypeFromString(xml_string):
  return atom.CreateClassFromXMLString(Snippettype, xml_string)

class Thumbnail(PhotosBaseElement):
  """The Google Photo `Thumbnail' element

  Used to display user's photo thumbnail (hackergotchi).
  
  (Not to be confused with the <media:thumbnail> element, which gives you
  small versions of the photo object.)"""
  
  _tag = 'thumbnail'
def ThumbnailFromString(xml_string):
  return atom.CreateClassFromXMLString(Thumbnail, xml_string)

class Timestamp(PhotosBaseElement):
  """The Google Photo `Timestamp' element
  Represented as the number of milliseconds since January 1st, 1970.
  
  
  Take a look at the convenience methods .isoformat() and .datetime():

  photo_epoch     = Time.text        # 1180294337000
  photo_isostring = Time.isoformat() # '2007-05-27T19:32:17.000Z'

  Alternatively: 
  photo_datetime  = Time.datetime()  # (requires python >= 2.3)
  """
  
  _tag = 'timestamp'
  def isoformat(self):
    """(string) Return the timestamp as a ISO 8601 formatted string,
    e.g. '2007-05-27T19:32:17.000Z'
    """
    import time
    epoch = float(self.text)/1000
    return time.strftime('%Y-%m-%dT%H:%M:%S.000Z', time.gmtime(epoch))
  
  def datetime(self):
    """(datetime.datetime) Return the timestamp as a datetime.datetime object

    Requires python 2.3
    """
    import datetime
    epoch = float(self.text)/1000
    return datetime.datetime.fromtimestamp(epoch)
def TimestampFromString(xml_string):
  return atom.CreateClassFromXMLString(Timestamp, xml_string)

class Truncated(PhotosBaseElement):
  """The Google Photo `Truncated' element

  You'll find this element in searches -- that is, feeds that combine the 
  `kind=photo' and `q=yoursearch' parameters in the request.

  See also gphoto:snippet and gphoto:snippettype.
  
  Possible values and their interpretation:
  0 -- unknown 
  """
  
  _tag = 'Truncated'
def TruncatedFromString(xml_string):
  return atom.CreateClassFromXMLString(Truncated, xml_string)

class User(PhotosBaseElement):
  "The Google Photo `User' element"
  
  _tag = 'user'
def UserFromString(xml_string):
  return atom.CreateClassFromXMLString(User, xml_string)

class Version(PhotosBaseElement):
  "The Google Photo `Version' element"
  
  _tag = 'version'
def VersionFromString(xml_string):
  return atom.CreateClassFromXMLString(Version, xml_string)

class Width(PhotosBaseElement):
  "The Google Photo `Width' element"
  
  _tag = 'width'
def WidthFromString(xml_string):
  return atom.CreateClassFromXMLString(Width, xml_string)

class Weight(PhotosBaseElement):
  """The Google Photo `Weight' element.

  The weight of the tag is the number of times the tag
  appears in the collection of tags currently being viewed.
  The default weight is 1, in which case this tags is omitted."""
  _tag = 'weight'
def WeightFromString(xml_string):
  return atom.CreateClassFromXMLString(Weight, xml_string)

class CommentAuthor(atom.Author):
  """The Atom `Author' element in CommentEntry entries is augmented to
  contain elements from the PHOTOS_NAMESPACE

  http://groups.google.com/group/Google-Picasa-Data-API/msg/819b0025b5ff5e38
  """
  _children = atom.Author._children.copy()
  _children['{%s}user' % PHOTOS_NAMESPACE] = ('user', User)
  _children['{%s}nickname' % PHOTOS_NAMESPACE] = ('nickname', Nickname)
  _children['{%s}thumbnail' % PHOTOS_NAMESPACE] = ('thumbnail', Thumbnail)
def CommentAuthorFromString(xml_string):
  return atom.CreateClassFromXMLString(CommentAuthor, xml_string)

########################## ################################

class AlbumData(object):
  _children = {}
  _children['{%s}id' % PHOTOS_NAMESPACE] = ('gphoto_id', Id) 
  _children['{%s}name' % PHOTOS_NAMESPACE] = ('name', Name)
  _children['{%s}location' % PHOTOS_NAMESPACE] = ('location', Location)
  _children['{%s}access' % PHOTOS_NAMESPACE] = ('access', Access)
  _children['{%s}bytesUsed' % PHOTOS_NAMESPACE] = ('bytesUsed', BytesUsed)
  _children['{%s}timestamp' % PHOTOS_NAMESPACE] = ('timestamp', Timestamp)
  _children['{%s}numphotos' % PHOTOS_NAMESPACE] = ('numphotos', Numphotos)
  _children['{%s}numphotosremaining' % PHOTOS_NAMESPACE] = \
    ('numphotosremaining', Numphotosremaining)
  _children['{%s}user' % PHOTOS_NAMESPACE] = ('user', User)
  _children['{%s}nickname' % PHOTOS_NAMESPACE] = ('nickname', Nickname)
  _children['{%s}commentingEnabled' % PHOTOS_NAMESPACE] = \
    ('commentingEnabled', CommentingEnabled)
  _children['{%s}commentCount' % PHOTOS_NAMESPACE] = \
    ('commentCount', CommentCount)
  ## NOTE: storing media:group as self.media, to create a self-explaining api
  gphoto_id = None
  name = None
  location = None
  access = None
  bytesUsed = None
  timestamp = None
  numphotos = None
  numphotosremaining = None
  user = None
  nickname = None
  commentingEnabled = None
  commentCount = None

class AlbumEntry(GPhotosBaseEntry, AlbumData):
  """All metadata for a Google Photos Album

  Take a look at AlbumData for metadata accessible as attributes to this object.

  Notes:
    To avoid name clashes, and to create a more sensible api, some
    objects have names that differ from the original elements:
  
    o media:group -> self.media,
    o geo:where -> self.geo,
    o photo:id -> self.gphoto_id
  """
  
  _kind = 'album'
  _children = GPhotosBaseEntry._children.copy()
  _children.update(AlbumData._children.copy())
  # child tags only for Album entries, not feeds
  _children['{%s}where' % GEORSS_NAMESPACE] = ('geo', Geo.Where)
  _children['{%s}group' % MEDIA_NAMESPACE] = ('media', Media.Group)
  media = Media.Group()
  geo = Geo.Where()
  
  def __init__(self, author=None, category=None, content=None,
      atom_id=None, link=None, published=None,
      title=None, updated=None,
      #GPHOTO NAMESPACE:
      gphoto_id=None, name=None, location=None, access=None, 
      timestamp=None, numphotos=None, user=None, nickname=None,
      commentingEnabled=None, commentCount=None, thumbnail=None,
      # MEDIA NAMESPACE:
      media=None,
      # GEORSS NAMESPACE:
      geo=None,
      extended_property=None,
      extension_elements=None, extension_attributes=None, text=None):
    GPhotosBaseEntry.__init__(self, author=author, category=category,
                        content=content, atom_id=atom_id, link=link,
                        published=published, title=title,
                        updated=updated, text=text,
                        extension_elements=extension_elements,
                        extension_attributes=extension_attributes)

    ## NOTE: storing photo:id as self.gphoto_id, to avoid name clash with atom:id
    self.gphoto_id = gphoto_id 
    self.name = name
    self.location = location
    self.access = access
    self.timestamp = timestamp
    self.numphotos = numphotos
    self.user = user
    self.nickname = nickname
    self.commentingEnabled = commentingEnabled
    self.commentCount = commentCount
    self.thumbnail = thumbnail
    self.extended_property = extended_property or []
    self.text = text
    ## NOTE: storing media:group as self.media, and geo:where as geo,
    ## to create a self-explaining api
    self.media = media or Media.Group()
    self.geo = geo or Geo.Where()

  def GetAlbumId(self):
    "Return the id of this album"
    
    return self.GetFeedLink().href.split('/')[-1]
          
  def GetPhotosUri(self):
    "(string) Return the uri to this albums feed of the PhotoEntry kind"
    return self._feedUri('photo')
  
  def GetCommentsUri(self):
    "(string) Return the uri to this albums feed of the CommentEntry kind"
    return self._feedUri('comment')
  
  def GetTagsUri(self):
    "(string) Return the uri to this albums feed of the TagEntry kind"
    return self._feedUri('tag')

def AlbumEntryFromString(xml_string):
  return atom.CreateClassFromXMLString(AlbumEntry, xml_string)
  
class AlbumFeed(GPhotosBaseFeed, AlbumData):
  """All metadata for a Google Photos Album, including its sub-elements
  
  This feed represents an album as the container for other objects.

  A Album feed contains entries of
  PhotoEntry, CommentEntry or TagEntry,
  depending on the `kind' parameter in the original query.

  Take a look at AlbumData for accessible attributes.
  
  """
  
  _children = GPhotosBaseFeed._children.copy()
  _children.update(AlbumData._children.copy())

  def GetPhotosUri(self):
    "(string) Return the uri to the same feed, but of the PhotoEntry kind"
    
    return self._feedUri('photo')
         
  def GetTagsUri(self):
    "(string) Return the uri to the same feed, but of the TagEntry kind"

    return self._feedUri('tag')
    
  def GetCommentsUri(self):
    "(string) Return the uri to the same feed, but of the CommentEntry kind"

    return self._feedUri('comment')
  
def AlbumFeedFromString(xml_string):
  return atom.CreateClassFromXMLString(AlbumFeed, xml_string)


class PhotoData(object):
  _children = {}
  ## NOTE: storing photo:id as self.gphoto_id, to avoid name clash with atom:id
  _children['{%s}id' % PHOTOS_NAMESPACE] = ('gphoto_id', Id) 
  _children['{%s}albumid' % PHOTOS_NAMESPACE] = ('albumid', Albumid)
  _children['{%s}checksum' % PHOTOS_NAMESPACE] = ('checksum', Checksum)
  _children['{%s}client' % PHOTOS_NAMESPACE] = ('client', Client)
  _children['{%s}height' % PHOTOS_NAMESPACE] = ('height', Height)
  _children['{%s}position' % PHOTOS_NAMESPACE] = ('position', Position)
  _children['{%s}rotation' % PHOTOS_NAMESPACE] = ('rotation', Rotation)
  _children['{%s}size' % PHOTOS_NAMESPACE] = ('size', Size)
  _children['{%s}timestamp' % PHOTOS_NAMESPACE] = ('timestamp', Timestamp)
  _children['{%s}version' % PHOTOS_NAMESPACE] = ('version', Version)
  _children['{%s}width' % PHOTOS_NAMESPACE] = ('width', Width)
  _children['{%s}commentingEnabled' % PHOTOS_NAMESPACE] = \
    ('commentingEnabled', CommentingEnabled)
  _children['{%s}commentCount' % PHOTOS_NAMESPACE] = \
    ('commentCount', CommentCount)
  ## NOTE: storing media:group as self.media, exif:tags as self.exif, and
  ## geo:where as self.geo, to create a self-explaining api
  _children['{%s}tags' % EXIF_NAMESPACE] = ('exif', Exif.Tags)
  _children['{%s}where' % GEORSS_NAMESPACE] = ('geo', Geo.Where)
  _children['{%s}group' % MEDIA_NAMESPACE] = ('media', Media.Group)
  # These elements show up in search feeds 
  _children['{%s}snippet' % PHOTOS_NAMESPACE] = ('snippet', Snippet)
  _children['{%s}snippettype' % PHOTOS_NAMESPACE] = ('snippettype', Snippettype)
  _children['{%s}truncated' % PHOTOS_NAMESPACE] = ('truncated', Truncated)
  gphoto_id = None
  albumid = None
  checksum = None
  client = None
  height = None
  position = None
  rotation = None
  size = None
  timestamp = None
  version = None
  width = None
  commentingEnabled = None
  commentCount = None
  snippet=None
  snippettype=None
  truncated=None
  media = Media.Group()
  geo = Geo.Where()
  tags = Exif.Tags()

class PhotoEntry(GPhotosBaseEntry, PhotoData):
  """All metadata for a Google Photos Photo

  Take a look at PhotoData for metadata accessible as attributes to this object.

  Notes:
    To avoid name clashes, and to create a more sensible api, some
    objects have names that differ from the original elements:
  
    o media:group -> self.media,
    o exif:tags -> self.exif,
    o geo:where -> self.geo,
    o photo:id -> self.gphoto_id
  """

  _kind = 'photo'
  _children = GPhotosBaseEntry._children.copy()
  _children.update(PhotoData._children.copy())
  
  def __init__(self, author=None, category=None, content=None,
      atom_id=None, link=None, published=None, 
      title=None, updated=None, text=None,
      # GPHOTO NAMESPACE:
      gphoto_id=None, albumid=None, checksum=None, client=None, height=None,
      position=None, rotation=None, size=None, timestamp=None, version=None,
      width=None, commentCount=None, commentingEnabled=None,
      # MEDIARSS NAMESPACE:
      media=None,
      # EXIF_NAMESPACE:
      exif=None,
      # GEORSS NAMESPACE:
      geo=None,
      extension_elements=None, extension_attributes=None):
    GPhotosBaseEntry.__init__(self, author=author, category=category,
                              content=content,
                              atom_id=atom_id, link=link, published=published,
                              title=title, updated=updated, text=text,
                              extension_elements=extension_elements,
                              extension_attributes=extension_attributes)
                              

    ## NOTE: storing photo:id as self.gphoto_id, to avoid name clash with atom:id
    self.gphoto_id = gphoto_id
    self.albumid = albumid
    self.checksum = checksum
    self.client = client
    self.height = height
    self.position = position
    self.rotation = rotation
    self.size = size
    self.timestamp = timestamp
    self.version = version
    self.width = width
    self.commentingEnabled = commentingEnabled
    self.commentCount = commentCount
    ## NOTE: storing media:group as self.media, to create a self-explaining api
    self.media = media or Media.Group()
    self.exif = exif or Exif.Tags()
    self.geo = geo or Geo.Where()

  def GetPostLink(self):
    "Return the uri to this photo's `POST' link (use it for updates of the object)"

    return self.GetFeedLink()

  def GetCommentsUri(self):
    "Return the uri to this photo's feed of CommentEntry comments"
    return self._feedUri('comment')    

  def GetTagsUri(self):
    "Return the uri to this photo's feed of TagEntry tags"
    return self._feedUri('tag')    

  def GetAlbumUri(self):
    """Return the uri to the AlbumEntry containing this photo"""

    href = self.GetSelfLink().href
    return href[:href.find('/photoid')]

def PhotoEntryFromString(xml_string):
  return atom.CreateClassFromXMLString(PhotoEntry, xml_string)

class PhotoFeed(GPhotosBaseFeed, PhotoData):
  """All metadata for a Google Photos Photo, including its sub-elements
  
  This feed represents a photo as the container for other objects.

  A Photo feed contains entries of
  CommentEntry or TagEntry,
  depending on the `kind' parameter in the original query.

  Take a look at PhotoData for metadata accessible as attributes to this object.
  
  """
  _children = GPhotosBaseFeed._children.copy()
  _children.update(PhotoData._children.copy())

  def GetTagsUri(self):
    "(string) Return the uri to the same feed, but of the TagEntry kind"
    
    return self._feedUri('tag')
  
  def GetCommentsUri(self):
    "(string) Return the uri to the same feed, but of the CommentEntry kind"
    
    return self._feedUri('comment')

def PhotoFeedFromString(xml_string):
  return atom.CreateClassFromXMLString(PhotoFeed, xml_string)

class TagData(GPhotosBaseData):
  _children = {}
  _children['{%s}weight' % PHOTOS_NAMESPACE] = ('weight', Weight)
  weight=None

class TagEntry(GPhotosBaseEntry, TagData):
  """All metadata for a Google Photos Tag

  The actual tag is stored in the .title.text attribute

  """

  _kind = 'tag'
  _children = GPhotosBaseEntry._children.copy()
  _children.update(TagData._children.copy())

  def __init__(self, author=None, category=None, content=None,
               atom_id=None, link=None, published=None,
               title=None, updated=None,
               # GPHOTO NAMESPACE:
               weight=None,
               extended_property=None,
               extension_elements=None, extension_attributes=None, text=None):
    GPhotosBaseEntry.__init__(self, author=author, category=category,
                              content=content,
                              atom_id=atom_id, link=link, published=published,
                              title=title, updated=updated, text=text,
                              extension_elements=extension_elements,
                              extension_attributes=extension_attributes)
    
    self.weight = weight

  def GetAlbumUri(self):
    """Return the uri to the AlbumEntry containing this tag"""

    href = self.GetSelfLink().href
    pos = href.find('/photoid')
    if pos == -1:
      return None
    return href[:pos]

  def GetPhotoUri(self):
    """Return the uri to the PhotoEntry containing this tag"""

    href = self.GetSelfLink().href
    pos = href.find('/tag')
    if pos == -1:
      return None
    return href[:pos]

def TagEntryFromString(xml_string):
  return atom.CreateClassFromXMLString(TagEntry, xml_string)


class TagFeed(GPhotosBaseFeed, TagData):
  """All metadata for a Google Photos Tag, including its sub-elements"""
  
  _children = GPhotosBaseFeed._children.copy()
  _children.update(TagData._children.copy())
  
def TagFeedFromString(xml_string):
  return atom.CreateClassFromXMLString(TagFeed, xml_string)

class CommentData(GPhotosBaseData):
  _children = {}
  ## NOTE: storing photo:id as self.gphoto_id, to avoid name clash with atom:id
  _children['{%s}id' % PHOTOS_NAMESPACE] = ('gphoto_id', Id)
  _children['{%s}albumid' % PHOTOS_NAMESPACE] = ('albumid', Albumid)
  _children['{%s}photoid' % PHOTOS_NAMESPACE] = ('photoid', Photoid)
  _children['{%s}author' % atom.ATOM_NAMESPACE] = ('author', [CommentAuthor,])
  gphoto_id=None
  albumid=None
  photoid=None
  author=None

class CommentEntry(GPhotosBaseEntry, CommentData):
  """All metadata for a Google Photos Comment
   
  The comment is stored in the .content.text attribute,
  with a content type in .content.type.


  """
  
  _kind = 'comment'
  _children = GPhotosBaseEntry._children.copy()
  _children.update(CommentData._children.copy())
  def __init__(self, author=None, category=None, content=None,
               atom_id=None, link=None, published=None,
               title=None, updated=None,
               # GPHOTO NAMESPACE:
               gphoto_id=None, albumid=None, photoid=None,
               extended_property=None,
               extension_elements=None, extension_attributes=None, text=None):
    
    GPhotosBaseEntry.__init__(self, author=author, category=category,
                              content=content,
                              atom_id=atom_id, link=link, published=published,
                              title=title, updated=updated,
                              extension_elements=extension_elements,
                              extension_attributes=extension_attributes,
                              text=text)
    
    self.gphoto_id = gphoto_id
    self.albumid = albumid
    self.photoid = photoid

  def GetCommentId(self):
    """Return the globally unique id of this comment"""
    return self.GetSelfLink().href.split('/')[-1]

  def GetAlbumUri(self):
    """Return the uri to the AlbumEntry containing this comment"""

    href = self.GetSelfLink().href
    return href[:href.find('/photoid')]

  def GetPhotoUri(self):
    """Return the uri to the PhotoEntry containing this comment"""

    href = self.GetSelfLink().href
    return href[:href.find('/commentid')]

def CommentEntryFromString(xml_string):
  return atom.CreateClassFromXMLString(CommentEntry, xml_string)

class CommentFeed(GPhotosBaseFeed, CommentData):
  """All metadata for a Google Photos Comment, including its sub-elements"""
  
  _children = GPhotosBaseFeed._children.copy()
  _children.update(CommentData._children.copy())

def CommentFeedFromString(xml_string):
  return atom.CreateClassFromXMLString(CommentFeed, xml_string)

class UserData(GPhotosBaseData):
  _children = {}
  _children['{%s}maxPhotosPerAlbum' % PHOTOS_NAMESPACE] = ('maxPhotosPerAlbum', MaxPhotosPerAlbum)
  _children['{%s}nickname' % PHOTOS_NAMESPACE] = ('nickname', Nickname)
  _children['{%s}quotalimit' % PHOTOS_NAMESPACE] = ('quotalimit', Quotalimit)
  _children['{%s}quotacurrent' % PHOTOS_NAMESPACE] = ('quotacurrent', Quotacurrent)
  _children['{%s}thumbnail' % PHOTOS_NAMESPACE] = ('thumbnail', Thumbnail)
  _children['{%s}user' % PHOTOS_NAMESPACE] = ('user', User)
  _children['{%s}id' % PHOTOS_NAMESPACE] = ('gphoto_id', Id)   

  maxPhotosPerAlbum=None
  nickname=None
  quotalimit=None
  quotacurrent=None
  thumbnail=None
  user=None
  gphoto_id=None
  

class UserEntry(GPhotosBaseEntry, UserData):
  """All metadata for a Google Photos User

  This entry represents an album owner and all appropriate metadata.

  Take a look at at the attributes of the UserData for metadata available.
  """
  _children = GPhotosBaseEntry._children.copy()
  _children.update(UserData._children.copy())
  _kind = 'user'
  
  def __init__(self, author=None, category=None, content=None,
               atom_id=None, link=None, published=None,
               title=None, updated=None,
               # GPHOTO NAMESPACE:
               gphoto_id=None, maxPhotosPerAlbum=None, nickname=None, quotalimit=None,
               quotacurrent=None, thumbnail=None, user=None,
               extended_property=None,
               extension_elements=None, extension_attributes=None, text=None):
    
    GPhotosBaseEntry.__init__(self, author=author, category=category,
                              content=content,
                              atom_id=atom_id, link=link, published=published,
                              title=title, updated=updated,
                              extension_elements=extension_elements,
                              extension_attributes=extension_attributes,
                              text=text)
                              
    
    self.gphoto_id=gphoto_id
    self.maxPhotosPerAlbum=maxPhotosPerAlbum
    self.nickname=nickname
    self.quotalimit=quotalimit
    self.quotacurrent=quotacurrent
    self.thumbnail=thumbnail
    self.user=user

  def GetAlbumsUri(self):
    "(string) Return the uri to this user's feed of the AlbumEntry kind"
    return self._feedUri('album')
  
  def GetPhotosUri(self):
    "(string) Return the uri to this user's feed of the PhotoEntry kind"
    return self._feedUri('photo')
  
  def GetCommentsUri(self):
    "(string) Return the uri to this user's feed of the CommentEntry kind"
    return self._feedUri('comment')
  
  def GetTagsUri(self):
    "(string) Return the uri to this user's feed of the TagEntry kind"
    return self._feedUri('tag')

def UserEntryFromString(xml_string):
  return atom.CreateClassFromXMLString(UserEntry, xml_string)
    
class UserFeed(GPhotosBaseFeed, UserData):
  """Feed for a User in the google photos api.

  This feed represents a user as the container for other objects.

  A User feed contains entries of
  AlbumEntry, PhotoEntry, CommentEntry, UserEntry or TagEntry,
  depending on the `kind' parameter in the original query.

  The user feed itself also contains all of the metadata available
  as part of a UserData object."""
  _children = GPhotosBaseFeed._children.copy()
  _children.update(UserData._children.copy())

  def GetAlbumsUri(self):
    """Get the uri to this feed, but with entries of the AlbumEntry kind."""
    return self._feedUri('album')

  def GetTagsUri(self):
    """Get the uri to this feed, but with entries of the TagEntry kind."""
    return self._feedUri('tag')

  def GetPhotosUri(self):
    """Get the uri to this feed, but with entries of the PhotosEntry kind."""
    return self._feedUri('photo')

  def GetCommentsUri(self):
    """Get the uri to this feed, but with entries of the CommentsEntry kind."""
    return self._feedUri('comment')

def UserFeedFromString(xml_string):
  return atom.CreateClassFromXMLString(UserFeed, xml_string)
  

  
def AnyFeedFromString(xml_string):
  """Creates an instance of the appropriate feed class from the
    xml string contents.

  Args:
    xml_string: str A string which contains valid XML. The root element
        of the XML string should match the tag and namespace of the desired
        class.

  Returns:
    An instance of the target class with members assigned according to the
    contents of the XML - or a basic gdata.GDataFeed instance if it is
    impossible to determine the appropriate class (look for extra elements
    in GDataFeed's .FindExtensions() and extension_elements[] ).
  """
  tree = ElementTree.fromstring(xml_string)
  category = tree.find('{%s}category' % atom.ATOM_NAMESPACE)
  if category is None:
    # TODO: is this the best way to handle this?
    return atom._CreateClassFromElementTree(GPhotosBaseFeed, tree)
  namespace, kind = category.get('term').split('#')
  if namespace != PHOTOS_NAMESPACE:
    # TODO: is this the best way to handle this?
    return atom._CreateClassFromElementTree(GPhotosBaseFeed, tree)
  ## TODO: is getattr safe this way?
  feed_class = getattr(gdata.photos, '%sFeed' % kind.title())
  return atom._CreateClassFromElementTree(feed_class, tree)

def AnyEntryFromString(xml_string):
  """Creates an instance of the appropriate entry class from the
    xml string contents.

  Args:
    xml_string: str A string which contains valid XML. The root element
        of the XML string should match the tag and namespace of the desired
        class.

  Returns:
    An instance of the target class with members assigned according to the
    contents of the XML - or a basic gdata.GDataEndry instance if it is
    impossible to determine the appropriate class (look for extra elements
    in GDataEntry's .FindExtensions() and extension_elements[] ).
  """
  tree = ElementTree.fromstring(xml_string)
  category = tree.find('{%s}category' % atom.ATOM_NAMESPACE)
  if category is None:
    # TODO: is this the best way to handle this?
    return atom._CreateClassFromElementTree(GPhotosBaseEntry, tree)
  namespace, kind = category.get('term').split('#')
  if namespace != PHOTOS_NAMESPACE:
    # TODO: is this the best way to handle this?
    return atom._CreateClassFromElementTree(GPhotosBaseEntry, tree)
  ## TODO: is getattr safe this way?
  feed_class = getattr(gdata.photos, '%sEntry' % kind.title())
  return atom._CreateClassFromElementTree(feed_class, tree)




"""Classes for reading/writing binary data (such as TLS records)."""

from compat import *

class Writer:
    def __init__(self, length=0):
        #If length is zero, then this is just a "trial run" to determine length
        self.index = 0
        self.bytes = createByteArrayZeros(length)

    def add(self, x, length):
        if self.bytes:
            newIndex = self.index+length-1
            while newIndex >= self.index:
                self.bytes[newIndex] = x & 0xFF
                x >>= 8
                newIndex -= 1
        self.index += length

    def addFixSeq(self, seq, length):
        if self.bytes:
            for e in seq:
                self.add(e, length)
        else:
            self.index += len(seq)*length

    def addVarSeq(self, seq, length, lengthLength):
        if self.bytes:
            self.add(len(seq)*length, lengthLength)
            for e in seq:
                self.add(e, length)
        else:
            self.index += lengthLength + (len(seq)*length)


class Parser:
    def __init__(self, bytes):
        self.bytes = bytes
        self.index = 0

    def get(self, length):
        if self.index + length > len(self.bytes):
            raise SyntaxError()
        x = 0
        for count in range(length):
            x <<= 8
            x |= self.bytes[self.index]
            self.index += 1
        return x

    def getFixBytes(self, lengthBytes):
        bytes = self.bytes[self.index : self.index+lengthBytes]
        self.index += lengthBytes
        return bytes

    def getVarBytes(self, lengthLength):
        lengthBytes = self.get(lengthLength)
        return self.getFixBytes(lengthBytes)

    def getFixList(self, length, lengthList):
        l = [0] * lengthList
        for x in range(lengthList):
            l[x] = self.get(length)
        return l

    def getVarList(self, length, lengthLength):
        lengthList = self.get(lengthLength)
        if lengthList % length != 0:
            raise SyntaxError()
        lengthList = int(lengthList/length)
        l = [0] * lengthList
        for x in range(lengthList):
            l[x] = self.get(length)
        return l

    def startLengthCheck(self, lengthLength):
        self.lengthCheck = self.get(lengthLength)
        self.indexCheck = self.index

    def setLengthCheck(self, length):
        self.lengthCheck = length
        self.indexCheck = self.index

    def stopLengthCheck(self):
        if (self.index - self.indexCheck) != self.lengthCheck:
            raise SyntaxError()

    def atLengthCheck(self):
        if (self.index - self.indexCheck) < self.lengthCheck:
            return False
        elif (self.index - self.indexCheck) == self.lengthCheck:
            return True
        else:
            raise SyntaxError()


#!/usr/bin/python
#
# Copyright (C) 2008 Google, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Allow Google Apps domain administrators to manage groups, group members and group owners.

  GroupsService: Provides methods to manage groups, members and owners.
"""

__author__ = 'google-apps-apis@googlegroups.com'


import urllib
import gdata.apps
import gdata.apps.service
import gdata.service


API_VER = '2.0'
BASE_URL = '/a/feeds/group/' + API_VER + '/%s'
GROUP_MEMBER_URL = BASE_URL + '?member=%s'
GROUP_MEMBER_DIRECT_URL = GROUP_MEMBER_URL + '&directOnly=%s'
GROUP_ID_URL = BASE_URL + '/%s'
MEMBER_URL = BASE_URL + '/%s/member'
MEMBER_WITH_SUSPENDED_URL = MEMBER_URL + '?includeSuspendedUsers=%s'
MEMBER_ID_URL = MEMBER_URL + '/%s'
OWNER_URL = BASE_URL + '/%s/owner'
OWNER_WITH_SUSPENDED_URL = OWNER_URL + '?includeSuspendedUsers=%s'
OWNER_ID_URL = OWNER_URL + '/%s'

PERMISSION_OWNER = 'Owner'
PERMISSION_MEMBER = 'Member'
PERMISSION_DOMAIN = 'Domain'
PERMISSION_ANYONE = 'Anyone'


class GroupsService(gdata.apps.service.PropertyService):
  """Client for the Google Apps Groups service."""

  def _ServiceUrl(self, service_type, is_existed, group_id, member_id, owner_email,
                  direct_only=False, domain=None, suspended_users=False):
    if domain is None:
      domain = self.domain

    if service_type == 'group':
      if group_id != '' and is_existed:
        return GROUP_ID_URL % (domain, group_id)
      elif member_id != '':
        if direct_only:
          return GROUP_MEMBER_DIRECT_URL % (domain, urllib.quote_plus(member_id),
                                            self._Bool2Str(direct_only))
        else:
          return GROUP_MEMBER_URL % (domain, urllib.quote_plus(member_id))
      else:
        return BASE_URL % (domain)

    if service_type == 'member':
      if member_id != '' and is_existed:
        return MEMBER_ID_URL % (domain, group_id, urllib.quote_plus(member_id))
      elif suspended_users:
        return MEMBER_WITH_SUSPENDED_URL % (domain, group_id,
                                            self._Bool2Str(suspended_users))
      else:
        return MEMBER_URL % (domain, group_id)

    if service_type == 'owner':
      if owner_email != '' and is_existed:
        return OWNER_ID_URL % (domain, group_id, urllib.quote_plus(owner_email))
      elif suspended_users:
        return OWNER_WITH_SUSPENDED_URL % (domain, group_id,
                                           self._Bool2Str(suspended_users))
      else:
        return OWNER_URL % (domain, group_id)

  def _Bool2Str(self, b):
    if b is None:
      return None
    return str(b is True).lower()

  def _IsExisted(self, uri):
    try:
      self._GetProperties(uri)
      return True
    except gdata.apps.service.AppsForYourDomainException, e:
      if e.error_code == gdata.apps.service.ENTITY_DOES_NOT_EXIST:
        return False
      else:
        raise e

  def CreateGroup(self, group_id, group_name, description, email_permission):
    """Create a group.

    Args:
      group_id: The ID of the group (e.g. us-sales).
      group_name: The name of the group.
      description: A description of the group
      email_permission: The subscription permission of the group.

    Returns:
      A dict containing the result of the create operation.
    """
    uri = self._ServiceUrl('group', False, group_id, '', '')
    properties = {}
    properties['groupId'] = group_id
    properties['groupName'] = group_name
    properties['description'] = description
    properties['emailPermission'] = email_permission
    return self._PostProperties(uri, properties)

  def UpdateGroup(self, group_id, group_name, description, email_permission):
    """Update a group's name, description and/or permission.

    Args:
      group_id: The ID of the group (e.g. us-sales).
      group_name: The name of the group.
      description: A description of the group
      email_permission: The subscription permission of the group.

    Returns:
      A dict containing the result of the update operation.
    """
    uri = self._ServiceUrl('group', True, group_id, '', '')
    properties = {}
    properties['groupId'] = group_id
    properties['groupName'] = group_name
    properties['description'] = description
    properties['emailPermission'] = email_permission
    return self._PutProperties(uri, properties)

  def RetrieveGroup(self, group_id):
    """Retrieve a group based on its ID.

    Args:
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      A dict containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('group', True, group_id, '', '')
    return self._GetProperties(uri)

  def RetrieveAllGroups(self):
    """Retrieve all groups in the domain.

    Args:
      None

    Returns:
      A list containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('group', True, '', '', '')
    return self._GetPropertiesList(uri)

  def RetrievePageOfGroups(self, start_group=None):
    """Retrieve one page of groups in the domain.
    
    Args:
      start_group: The key to continue for pagination through all groups.
      
    Returns:
      A feed object containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('group', True, '', '', '')
    if start_group is not None:
      uri += "?start="+start_group
    property_feed = self._GetPropertyFeed(uri)
    return property_feed

  def RetrieveGroups(self, member_id, direct_only=False):
    """Retrieve all groups that belong to the given member_id.

    Args:
      member_id: The member's email address (e.g. member@example.com).
      direct_only: Boolean whether only return groups that this member directly belongs to.

    Returns:
      A list containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('group', True, '', member_id, '', direct_only=direct_only)
    return self._GetPropertiesList(uri)

  def DeleteGroup(self, group_id):
    """Delete a group based on its ID.

    Args:
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      A dict containing the result of the delete operation.
    """
    uri = self._ServiceUrl('group', True, group_id, '', '')
    return self._DeleteProperties(uri)

  def AddMemberToGroup(self, member_id, group_id):
    """Add a member to a group.

    Args:
      member_id: The member's email address (e.g. member@example.com).
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      A dict containing the result of the add operation.
    """
    uri = self._ServiceUrl('member', False, group_id, member_id, '')
    properties = {}
    properties['memberId'] = member_id
    return self._PostProperties(uri, properties)

  def IsMember(self, member_id, group_id):
    """Check whether the given member already exists in the given group.

    Args:
      member_id: The member's email address (e.g. member@example.com).
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      True if the member exists in the group.  False otherwise.
    """
    uri = self._ServiceUrl('member', True, group_id, member_id, '')
    return self._IsExisted(uri)

  def RetrieveMember(self, member_id, group_id):
    """Retrieve the given member in the given group.

    Args:
      member_id: The member's email address (e.g. member@example.com).
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      A dict containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('member', True, group_id, member_id, '')
    return self._GetProperties(uri)

  def RetrieveAllMembers(self, group_id, suspended_users=False):
    """Retrieve all members in the given group.

    Args:
      group_id: The ID of the group (e.g. us-sales).
      suspended_users: A boolean; should we include any suspended users in
        the membership list returned?

    Returns:
      A list containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('member', True, group_id, '', '',
                           suspended_users=suspended_users)
    return self._GetPropertiesList(uri)
    
  def RetrievePageOfMembers(self, group_id, suspended_users=False, start=None):
    """Retrieve one page of members of a given group.
    
    Args:
      group_id: The ID of the group (e.g. us-sales).
      suspended_users: A boolean; should we include any suspended users in
        the membership list returned?
      start: The key to continue for pagination through all members.

    Returns:
      A feed object containing the result of the retrieve operation.
    """
    
    uri = self._ServiceUrl('member', True, group_id, '', '',
                           suspended_users=suspended_users)
    if start is not None:
      if suspended_users:
        uri += "&start="+start
      else:
        uri += "?start="+start
    property_feed = self._GetPropertyFeed(uri)
    return property_feed

  def RemoveMemberFromGroup(self, member_id, group_id):
    """Remove the given member from the given group.

    Args:
      member_id: The member's email address (e.g. member@example.com).
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      A dict containing the result of the remove operation.
    """
    uri = self._ServiceUrl('member', True, group_id, member_id, '')
    return self._DeleteProperties(uri)

  def AddOwnerToGroup(self, owner_email, group_id):
    """Add an owner to a group.

    Args:
      owner_email: The email address of a group owner.
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      A dict containing the result of the add operation.
    """
    uri = self._ServiceUrl('owner', False, group_id, '', owner_email)
    properties = {}
    properties['email'] = owner_email
    return self._PostProperties(uri, properties)

  def IsOwner(self, owner_email, group_id):
    """Check whether the given member an owner of the given group.

    Args:
      owner_email: The email address of a group owner.
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      True if the member is an owner of the given group.  False otherwise.
    """
    uri = self._ServiceUrl('owner', True, group_id, '', owner_email)
    return self._IsExisted(uri)

  def RetrieveOwner(self, owner_email, group_id):
    """Retrieve the given owner in the given group.

    Args:
      owner_email: The email address of a group owner.
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      A dict containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('owner', True, group_id, '', owner_email)
    return self._GetProperties(uri)

  def RetrieveAllOwners(self, group_id, suspended_users=False):
    """Retrieve all owners of the given group.

    Args:
      group_id: The ID of the group (e.g. us-sales).
      suspended_users: A boolean; should we include any suspended users in
        the ownership list returned?

    Returns:
      A list containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('owner', True, group_id, '', '',
                           suspended_users=suspended_users)
    return self._GetPropertiesList(uri)
    
  def RetrievePageOfOwners(self, group_id, suspended_users=False, start=None):
    """Retrieve one page of owners of the given group.
    
    Args:
      group_id: The ID of the group (e.g. us-sales).
      suspended_users: A boolean; should we include any suspended users in
        the ownership list returned?
      start: The key to continue for pagination through all owners.

    Returns:
      A feed object containing the result of the retrieve operation.
    """
    uri = self._ServiceUrl('owner', True, group_id, '', '',
                           suspended_users=suspended_users)
    if start is not None:
      if suspended_users:
        uri += "&start="+start
      else:
        uri += "?start="+start
    property_feed = self._GetPropertyFeed(uri)
    return property_feed
        
  def RemoveOwnerFromGroup(self, owner_email, group_id):
    """Remove the given owner from the given group.

    Args:
      owner_email: The email address of a group owner.
      group_id: The ID of the group (e.g. us-sales).

    Returns:
      A dict containing the result of the remove operation.
    """
    uri = self._ServiceUrl('owner', True, group_id, '', owner_email)
    return self._DeleteProperties(uri)



__doc__ = """
This module implements a subset of the XPath standard:
- tags
- indices
- attributes
- descendants

This was created because I needed a pure Python XPath parser.

Generally XPath solutions will normalize the HTML into XHTML before selecting nodes.
However this module tries to navigate the HTML structure directly without normalizing by searching for the next closing tag.
"""

#TODO:
# - parent
# - search by text: text() == '...'
# - return xpath for most similar to text

import re
import sys
import urllib2
import urlparse
from optparse import OptionParser
import adt
import common
import settings


class Doc:
    """Wrapper around a parsed webpage

    html:
        The content of webpage to parse
    remove:
        A list of tags to remove

    >>> doc = Doc('<div>abc<a class="link">LINK 1</a><div><a>LINK 2</a>def</div>abc</div>ghi<div><a>LINK 3</a>jkl</div>')
    >>> doc.search('/div/a')
    ['LINK 1', 'LINK 3']
    >>> doc.search('/div/a[@class="link"]')
    ['LINK 1']
    >>> doc.search('/div[1]//a')
    ['LINK 1', 'LINK 2']
    >>> doc.search('/div/a/@class')
    ['link', '']
    >>> doc.search('/div[-1]/a')
    ['LINK 3']

    >>> # test searching unicode
    >>> doc = Doc(u'<a href="http://www.google.com" class="flink">google</a>')
    >>> doc.get('//a[@class="flink"]')
    u'google'

    >>> # test finding just the first instance for a large amount of content
    >>> doc = Doc('<div><span>content</span></div>' * 10000)
    >>> doc.get('//span')
    'content'

    >>> # test extracting attribute of self closing tag
    >>> Doc('<div><img src="img.png"></div>').get('/div/img/@src')
    'img.png'

    >>> # test extracting attribute after self closing tag
    >>> Doc('<div><br><p>content</p></div>').get('/div/p')
    'content'
    """

    # regex to find a tag
    _tag_regex = re.compile('<([\w\:]+)')
    # regex to find an attribute
    _attributes_regex = re.compile('([\w\:-]+)\s*=\s*(".*?"|\'.*?\'|\S+)', re.DOTALL)
    # regex to find content of a tag
    _content_regex = re.compile('<.*?>(.*)</.*?>$', re.DOTALL)


    def __init__(self, html, remove=None):
        self.html = self._clean(html, remove)
        self.num_searches = 0

    def get(self, xpath):
        """Return the first result from this XPath selection
        """
        results = self._xpath(self.parse(xpath), self.html, limit=1)
        return common.first(results)

    def search(self, xpath):
        """Return all results from this XPath selection
        """
        return self._xpath(self.parse(xpath), self.html, limit=sys.maxint)


    def _xpath(self, path, html, limit):
        """Recursively search HTML for content at XPath
        """
        counter, separator, tag, index, attributes = path.pop(0)
        if counter == 0:
            self.num_searches += 1

        results = []
        if tag == '..':
            # parent
            raise common.WebScrapingError('.. not yet supported')
            results.append(self.get_parent(html))
        elif tag == 'text()':
            # extract child text
            text = self._get_content(self._get_html(html))
            results.append(common.remove_tags(text, keep_children=False))
            # check if next tag is selecting attribute
        elif tag.startswith('@'):
            attr = tag[1:].lower()
            #parent = self.get_parent(context)
            value = self._get_attributes(html).get(attr, '')
            results.append(value)
        else:
            # have tag
            if counter > 0:
                # get child html when not at root
                html = self._get_content(html)

            # search direct children if / and all descendants if //
            search_fn = self._find_children if separator == '' else self._find_descendants
            matches = search_fn(html, tag)

            # support negative indices
            if index is not None and index < 0:
                matches = list(matches)
                index += len(matches) + 1

            for child_i, child in enumerate(matches):
                # check if matches index
                if index is None or index == child_i + 1:
                    # check if matches attributes
                    if not attributes or self._match_attributes(attributes, self._get_attributes(child)):
                        if path:
                            results.extend(self._xpath(path[:], child, limit))
                        else:
                            # final node
                            results.append(self._get_content(child))
                        if len(results) > limit:
                            break

            #if not children:
            #    attributes_s = attributes and ''.join('[@%s="%s"]' % a for a in attributes) or ''
            #    common.logger.debug('No matches for <%s%s%s> (tag %d)' % (tag, index and '[%d]' % index or '', attributes_s, tag_i + 1))
        return results



    def _clean(self, html, remove):
        """Remove specified unhelpful tags and comments
        """
        self.remove = remove
        html = re.compile('<!--.*?-->', re.DOTALL).sub('', html) # remove comments
        if remove:
            # XXX combine tag list into single regex, if can match same at start and end
            for tag in remove:
                html = re.compile('<' + tag + '[^>]*?/>', re.DOTALL | re.IGNORECASE).sub('', html)
                html = re.compile('<' + tag + '[^>]*?>.*?</' + tag + '>', re.DOTALL | re.IGNORECASE).sub('', html)
                html = re.compile('<' + tag + '[^>]*?>', re.DOTALL | re.IGNORECASE).sub('', html)
        return html


    def parse(self, xpath):
        """Parse the xpath into: counter, separator, tag, index, and attributes

        >>> doc = Doc('')
        >>> doc.parse('/div[1]//span[@class="text"]')
        [(0, '', 'div', 1, []), (1, '/', 'span', None, [('class', 'text')])]
        >>> doc.parse('//li[-2]')
        [(0, '/', 'li', -2, [])]
        >>> doc.parse('/div[@id="content"]//span[1][@class="text"][@title=""]/a')
        [(0, '', 'div', None, [('id', 'content')]), (1, '/', 'span', 1, [('class', 'text'), ('title', '')]), (2, '', 'a', None, [])]
        """
        tokens = []
        counter = 0
        for separator, token in re.compile('(|/|\.\.)/([^/]+)').findall(xpath):
            index, attributes = None, []
            if '[' in token:
                tag = token[:token.find('[')]
                for attribute in re.compile('\[(.*?)\]').findall(token):
                    try:
                        index = int(attribute)
                    except ValueError:
                        match = re.compile('@(.*?)=["\']?(.*?)["\']?$').search(attribute)
                        if match:
                            key, value = match.groups()
                            attributes.append((key.lower(), value.lower()))
                        else:
                            raise common.WebScrapingError('Unknown format: ' + attribute)
            else:
                tag = token
            tokens.append((counter, separator, tag, index, attributes))
            counter += 1
        return tokens


    def _get_attributes(self, html):
        """Extract the attributes of the passed HTML tag

        >>> doc = Doc('')
        >>> doc._get_attributes('<div id="ID" name="MY NAME" max-width="20" class=abc>content <span class="inner name">SPAN</span></div>')
        {'max-width': '20', 'class': 'abc', 'id': 'ID', 'name': 'MY NAME'}
        """

        for i, c in enumerate(html):
            if c == '>':
                html = html[:i]
                break
        return dict((name.lower().strip(), value.strip('\'" ')) for (name, value) in Doc._attributes_regex.findall(html))


    def _match_attributes(self, desired_attributes, available_attributes):
        """Returns True if all of desired attributes are in available attributes
        Supports regex, which is not part of the XPath standard but is so useful!

        >>> doc = Doc('')
        >>> doc._match_attributes([], {})
        True
        >>> doc._match_attributes([('class', 'test')], {})
        False
        >>> doc._match_attributes([], {'id':'test', 'class':'test2'})
        True
        >>> doc._match_attributes([('class', 'test')], {'id':'test', 'class':'test2'})
        False
        >>> doc._match_attributes([('class', 'test')], {'id':'test2', 'class':'test'})
        True
        >>> doc._match_attributes([('class', 'test'), ('id', 'content')], {'id':'test', 'class':'content'})
        False
        >>> doc._match_attributes([('class', 'test'), ('id', 'content')], {'id':'content', 'class':'test'})
        True
        >>> doc._match_attributes([('class', 'test\d')], {'id':'test', 'class':'test2'})
        True
        >>> doc._match_attributes([('class', 'test\d')], {'id':'test2', 'class':'test'})
        False
        """
        for name, value in desired_attributes:
            if name not in available_attributes or not re.match(re.compile(value + '$', re.IGNORECASE), available_attributes[name]):
                return False
        return True


    def _get_html(self, context):
        """Return HTML at this context
        """
        return context
        if context:
            i, j = context
        else:
            i, j = 0, len(self.html)
        return self.html[i:j]


    def _get_content(self, context, default=''):
        """Extract the child HTML of a the passed HTML tag

        >>> doc = Doc('')
        >>> doc._get_content('<div id="ID" name="NAME">content <span>SPAN</span></div>')
        'content <span>SPAN</span>'
        """
        match = Doc._content_regex.match(self._get_html(context))
        if match:
            content = match.groups()[0]
        else:
            content = default
        return content


    def _find_children(self, html, tag):
        """Find children with this tag type

        >>> doc = Doc('')
        >>> list(doc._find_children('<span>1</span><div>abc<div>def</div>abc</div>ghi<div>jkl</div>', 'div'))
        ['<div>abc<div>def</div>abc</div>', '<div>jkl</div>']
        >>> list(doc._find_children('<tbody><tr><td></td></tr></tbody>', 'tbody'))
        ['<tbody><tr><td></td></tr></tbody>']
        >>> list(doc._find_children('<tr><td></td></tr>', 'tbody'))
        ['<tr><td></td></tr>']
        """
        found = True
        num_found = 0
        orig_html = html
        while found:
            html = self._jump_next_tag(html)
            if html:
                tag_html, html = self._split_tag(html)
                if tag_html:
                    if tag.lower() in ('*', self._get_tag(tag_html).lower()):
                        num_found += 1
                        yield tag_html
                else:
                    found = False
            else:
                found = False

        if tag == 'tbody' and num_found == 0:
            # skip tbody, which firefox includes in xpath when does not exist
            yield orig_html


    def _find_descendants(self, html, tag):
        """Find descendants with this tag type

        >>> doc = Doc('')
        >>> list(doc._find_descendants('<span>1</span><div>abc<div>def</div>abc</div>ghi<div>jkl</div>', 'div'))
        ['<div>abc<div>def</div>abc</div>', '<div>def</div>', '<div>jkl</div>']
        """
        # XXX search with attribute here
        if tag == '*':
            raise common.WebScrapingError("`*' not currently supported for //")
        for match in re.compile('<%s' % tag, re.DOTALL | re.IGNORECASE).finditer(html):
            tag_html = html[match.start():]
            tag_html, _ = self._split_tag(tag_html)
            yield tag_html


    def _jump_next_tag(self, html):
        """Return html at start of next tag

        >>> doc = Doc('')
        >>> doc._jump_next_tag('<div>abc</div>')
        '<div>abc</div>'
        >>> doc._jump_next_tag(' <div>abc</div>')
        '<div>abc</div>'
        >>> doc._jump_next_tag('</span> <div>abc</div>')
        '<div>abc</div>'
        >>> doc._jump_next_tag(' <br> <div>abc</div>')
        '<br> <div>abc</div>'
        """
        while 1:
            match = Doc._tag_regex.search(html)
            if match:
                return html[match.start():]
            else:
                return None


    def _get_tag(self, html):
        """Find tag type at this location

        >>> doc = Doc('')
        >>> doc._get_tag('<div>abc</div>')
        'div'
        >>> doc._get_tag(' <div>')
        >>> doc._get_tag('div')
        """
        match = Doc._tag_regex.match(html)
        if match:
            return match.groups()[0]
        else:
            return None


    def _split_tag(self, html):
        """Extract starting tag and contents from HTML

        >>> doc = Doc('')
        >>> doc._split_tag('<div>abc<div>def</div>abc</div>ghi<div>jkl</div>')
        ('<div>abc<div>def</div>abc</div>', 'ghi<div>jkl</div>')
        >>> doc._split_tag('<br /><div>abc</div>')
        ('<br />', '<div>abc</div>')
        >>> doc._split_tag('<div>abc<div>def</div>abc</span>')
        ('<div>abc<div>def</div>abc</span></div>', '')
        >>> # test efficiency of splits
        >>> a = [doc._split_tag('<div>abc<div>def</div>abc</span>') for i in range(10000)]
        """
        i = None
        tag = self._get_tag(html)
        depth = 0 # how far nested
        for match in re.compile('</?%s.*?>' % tag, re.DOTALL | re.IGNORECASE).finditer(html):
            if html[match.start() + 1] == '/':
                depth -= 1 # found closing tag
            elif tag in common.EMPTY_TAGS:
                pass # this tag type does not close
            elif html[match.end() - 2] == '/':
                pass # tag starts and ends (eg <br />)
            else:
                depth += 1 # found opening tag
            if depth == 0:
                # found top level match
                i = match.end()
                break
        if i is None:
            # all html is within this tag
            return html + '</%s>' % tag, ''
        else:
            return html[:i], html[i:]


    def _parent_tag(self, html):
        """Find parent tag of this current tag

        >> doc = Doc('<p><div><span id="abc">empty</span></div></p>')
        >> doc._parent_tag('<span id="abc">empty</span>')
        '<div><span id="abc">empty</span></div>'
        >> doc = Doc('<div><p></p><span id="abc">empty</span></div>')
        >> doc._parent_tag('<span id="abc">empty</span>')
        '<div><p></p><span id="abc">empty</span></div>'
        """
        raise Exception('Not implemented')
        #index = self.html.find(html)
        #while index >= 0:
        #    index = self.html.rfind('<', start=0, end=index)


try:
    from lxml import html as lxmlhtml
except ImportError:
    LxmlDoc = None
else:
    # is lxml is supported create wrapper
    class LxmlDoc:
        def __init__(self, html, **kwargs):
            self.doc = lxmlhtml.fromstring(html)

        def __eq__(self, html):
            return self.orig_html is html

        def get(self, path):
            es = self.doc.xpath(path)
            if es:
                return self.tostring(es[0])
            return ''

        def search(self, path):
            return [self.tostring(e) for e in self.doc.xpath(path)]

        def tostring(self, node):
            try:
                return ''.join(filter(None, 
                    [node.text] + [lxmlhtml.tostring(e) for e in node] + [node.tail] 
                ))
            except AttributeError:
                return node



prev_doc = {}
def _get_doc(html, remove):
    """Return previous doc object if same HTML, else create new one

    >>> html = '<div>1</div><div>2</div>'
    >>> get(html, '/div', None)
    '1'
    >>> search(html, '//div', None)
    ['1', '2']
    >>> _get_doc(html, None).num_searches
    2
    """
    global prev_doc
    key = html, remove
    doc = prev_doc.get(key)
    if doc:
        pass # can reuse current doc
    else:
        doc = Doc(html, remove=remove)
        prev_doc = {key: doc}
    return doc


def get(html, xpath, remove=('br', 'hr')):
    """Return first element from XPath search of HTML
    """
    return _get_doc(html, remove=remove).get(xpath)

def search(html, xpath, remove=('br', 'hr')):
    """Return all elements from XPath search of HTML
    """
    return _get_doc(html, remove=remove).search(xpath)

def find_children(html, tag, remove=None):
    """Find children with this tag type
    """
    return _get_doc(html, remove=remove)._find_children(html, tag)

js_re = re.compile('location.href ?= ?[\'"](.*?)[\'"]')
def get_links(html, url=None, local=True, external=True):
    """Return all links from html and convert relative to absolute if source url is provided

    html:
        HTML to parse
    url:
        optional URL for determining path of relative links
    local:
        whether to include links from same domain
    external:
        whether to include linkes from other domains
    """
    def normalize_link(link):
        if urlparse.urlsplit(link).scheme in ('http', 'https', ''):
            if '#' in link:
                link = link[:link.index('#')]
            if url:
                link = urlparse.urljoin(url, link)
                if not local and common.same_domain(url, link):
                    # local links not included
                    link = None
                if not external and not common.same_domain(url, link):
                    # external links not included
                    link = None
        else:
            link = None # ignore mailto, etc
        return link
    a_links = search(html, '//a/@href')
    js_links = js_re.findall(html)
    links = []
    for link in a_links + js_links:
        try:
            link = normalize_link(link)
        except UnicodeError:
            pass
        else:
            if link and link not in links:
                links.append(link)
    return links
